==> vmaid-1.9.1.tar.gz: vmaid-1.9.1/avicore/wave.c <==
/*
    avicore
    copyright (c) 1998-2005 Kazuki IWAMOTO http://www.maid.org/ iwm@maid.org

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "wave.h"


/******************************************************************************
*                                                                             *
* ja:PCM加工関数                                                              *
*                                                                             *
******************************************************************************/
typedef struct _WaveConvChannels
{
  guint16 channels;
  guint16 bits_per_sample;
} WaveConvChannels;


/*  ja:PCMのチャンネル変換を開始する
    wfx0,元のWaveFormatEx構造体へのポインタ
    wfx1,新しいWaveFormatEx構造体へのポインタ
     RET,変換ハンドル,NULL:変換不可能                                       */
WaveConvChannels *
wave_convert_channels_begin (WaveFormatEx *wfx0,
                             WaveFormatEx *wfx1)
{
  WaveConvChannels *wcc;

  wcc = g_malloc (sizeof (WaveConvChannels));
  wcc->channels = wfx_get_channels (wfx0);
  wcc->bits_per_sample = wfx_get_bits_per_sample (wfx0);
  return wcc;
}


/*  ja:PCMのチャンネル変換をする
            wcc,変換ハンドル
         in_buf,入力データ
     in_samples,入力データのサンプル数
        out_buf,出力データ
    out_samples,出力データのサンプル数
            RET,TRUE:正常終了,FALSE:エラー                                  */
gboolean
wave_convert_channels (WaveConvChannels *wcc,
                       gpointer          in_buf,
                       const gsize       in_samples,
                       gpointer         *out_buf,
                       gsize            *out_samples)
{
  gpointer result;
  gsize i;

  if (wcc->channels == 1)
    {
      if (wcc->bits_per_sample == 8)
        {
          result = g_malloc (in_samples * 2 * sizeof (guint8));
          for (i = 0; i < in_samples; i++)
            ((guint8 *)result)[i * 2] =
            ((guint8 *)result)[i * 2 + 1] = ((guint8 *)in_buf)[i];
        }
      else
        {
          result = g_malloc (in_samples * 2 * sizeof (gint16));
          for (i = 0; i < in_samples; i++)
            ((gint16 *)result)[i * 2] =
            ((gint16 *)result)[i * 2 + 1] = ((gint16 *)in_buf)[i];
        }
    }
  else
    {
      if (wcc->bits_per_sample == 8)
        {
          result = g_malloc (in_samples * sizeof (guint8));
          for (i = 0; i < in_samples; i++)
            ((guint8 *)result)[i] = (((guint8 *)in_buf)[i * 2]
                                        + ((guint8 *)in_buf)[i * 2 + 1]) / 2;
        }
      else
        {
          result = g_malloc (in_samples * sizeof (gint16));
          for (i = 0; i < in_samples; i++)
            ((gint16 *)result)[i] =
                    GINT16_TO_LE ((GINT16_FROM_LE (((gint16 *)in_buf)[i * 2])
                        + GINT16_FROM_LE (((gint16 *)in_buf)[i * 2 + 1])) / 2);
        }
    }
  if (out_buf)
    *out_buf = result;
  else
    g_free (result);
  if (out_samples)
    *out_samples = in_samples;
  return TRUE;
}


/*  ja:PCMのチャンネル変換を終了する
      wave_conv,変換ハンドル
        out_buf,出力データ
    out_samples,出力データのサンプル数
            RET,TRUE:正常終了,FALSE:エラー                                  */
gboolean
wave_convert_channels_end (WaveConvChannels *wcc,
                           gpointer         *out_buf,
                           gsize            *out_samples)
{
  if (out_buf)
    *out_buf = NULL;
  if (out_samples)
    *out_samples = 0;
  g_free (wcc);
  return TRUE;
}


typedef struct _WaveConvBitPerSample
{
  guint16 channels;
  guint16 bits_per_sample;
} WaveConvBitPerSample;


/*  ja:PCMのビット数変換を開始する
    wfx0,元のWaveFormatEx構造体へのポインタ
    wfx1,新しいWaveFormatEx構造体へのポインタ
     RET,変換ハンドル,NULL:変換不可能                                       */
WaveConvBitPerSample *
wave_convert_bits_per_sample_begin (WaveFormatEx *wfx0,
                                    WaveFormatEx *wfx1)
{
  WaveConvBitPerSample *wcb;

  wcb = g_malloc (sizeof (WaveConvBitPerSample));
  wcb->channels = wfx_get_channels (wfx0);
  wcb->bits_per_sample = wfx_get_bits_per_sample (wfx0);
  return wcb;
}


/*  ja:PCMのビット数変換をする
            wcb,変換ハンドル
         in_buf,入力データ
     in_samples,入力データのサンプル数
        out_buf,出力データ
    out_samples,出力データのサンプル数
            RET,TRUE:正常終了,FALSE:エラー                                  */
gboolean
wave_convert_bits_per_sample (WaveConvBitPerSample *wcb,
                              gpointer              in_buf,
                              const gsize           in_samples,
                              gpointer             *out_buf,
                              gsize                *out_samples)
{
  gpointer result;
  gsize i;

  if (wcb->bits_per_sample == 8)
    {
      result = g_malloc (in_samples * wcb->channels * sizeof (gint16));
      for (i = in_samples * wcb->channels; i > 0; i--)
        {
          gint c;

          c = ((guint8 *)in_buf)[i] - 128;
          c = c * 256 + (c % 2 == 0 ? 0 : 255);
          ((gint16 *)result)[i] = GINT16_TO_LE (c);
        }
    }
  else
    {
      result = g_malloc (in_samples * wcb->channels * sizeof (guint8));
      for (i = in_samples * wcb->channels; i > 0; i--)
        ((guint8 *)result)[i] = (GINT16_TO_LE (((gint16 *)in_buf)[i]) + 32768)
                                                                        / 256;
    }
  if (out_buf)
    *out_buf = result;
  else
    g_free (result);
  if (out_samples)
    *out_samples = in_samples;
  return TRUE;
}


/*  ja:PCMのビット数変換を終了する
      wave_conv,変換ハンドル
        out_buf,出力データ
    out_samples,出力データのサンプル数
            RET,TRUE:正常終了,FALSE:エラー                                  */
gboolean
wave_convert_bits_per_sample_end (WaveConvBitPerSample *wcb,
                                  gpointer             *out_buf,
                                  gsize                *out_samples)
{
  if (out_buf)
    *out_buf = NULL;
  if (out_samples)
    *out_samples = 0;
  g_free (wcb);
  return TRUE;
}


typedef struct _WaveConvGeneric *WaveConvGeneric;
typedef WaveConvGeneric (*WaveConvBeginFunc)(WaveFormatEx *wfx0,
                                             WaveFormatEx *wfx1);
typedef gboolean (*WaveConvFunc)(WaveConvGeneric  wch,
                                 gpointer         in_buf,
                                 gsize            in_samples,
                                 gpointer        *out_buf,
                                 gsize           *out_samples);
typedef gboolean (*WaveConvEndFunc)(WaveConvGeneric  wch,
                                    gpointer        *out_buf,
                                    gsize           *out_samples);
typedef struct _WaveConvProccess
{
  WaveConvBeginFunc func_begin;
  WaveConvFunc func;
  WaveConvEndFunc func_end;
  WaveConvGeneric handle;
} WaveConvProccess;
struct _WaveConv
{
  guint16 channels0;
  guint16 channels1;
  guint16 bits_per_sample0;
  guint16 bits_per_sample1;
  gint procs;
  WaveConvProccess *proc;
};


/*  ja:PCMの変換を開始する
    wfx0,元のWaveFormatEx構造体へのポインタ
    wfx1,新しいWaveFormatEx構造体へのポインタ
     RET,変換ハンドル,NULL:変換不可能                                       */
WaveConv *
wave_convert_begin (WaveFormatEx *wfx0,
                    WaveFormatEx *wfx1)
{
  gint i;
  WaveConv *wcv;

  wcv = g_malloc0 (sizeof (WaveConv));
  wcv->channels0 = wfx_get_channels (wfx0);
  wcv->channels1 = wfx_get_channels (wfx1);
  wcv->bits_per_sample0 = wfx_get_bits_per_sample (wfx0);
  wcv->bits_per_sample1 = wfx_get_bits_per_sample (wfx1);
  if (wcv->channels0 == 1 && wcv->channels1 == 2)
    {
      wcv->proc = g_realloc (wcv->proc,
                                (wcv->procs + 1) * sizeof (WaveConvProccess));
      wcv->proc[wcv->procs].func_begin
                            = (WaveConvBeginFunc)wave_convert_channels_begin;
      wcv->proc[wcv->procs].func = (WaveConvFunc)wave_convert_channels;
      wcv->proc[wcv->procs].func_end
                                = (WaveConvEndFunc)wave_convert_channels_end;
      wcv->procs++;
    }
  if (wcv->bits_per_sample0 == 8 && wcv->bits_per_sample1 == 16)
    {
      wcv->proc = g_realloc (wcv->proc,
                                (wcv->procs + 1) * sizeof (WaveConvProccess));
      wcv->proc[wcv->procs].func_begin
                    = (WaveConvBeginFunc)wave_convert_bits_per_sample_begin;
      wcv->proc[wcv->procs].func = (WaveConvFunc)wave_convert_bits_per_sample;
      wcv->proc[wcv->procs].func_end
                        = (WaveConvEndFunc)wave_convert_bits_per_sample_end;
      wcv->procs++;
    }
  if (wcv->bits_per_sample0 == 16 && wcv->bits_per_sample1 == 8)
    {
      wcv->proc = g_realloc (wcv->proc,
                                (wcv->procs + 1) * sizeof (WaveConvProccess));
      wcv->proc[wcv->procs].func_begin
                    = (WaveConvBeginFunc)wave_convert_bits_per_sample_begin;
      wcv->proc[wcv->procs].func = (WaveConvFunc)wave_convert_bits_per_sample;
      wcv->proc[wcv->procs].func_end
                        = (WaveConvEndFunc)wave_convert_bits_per_sample_end;
      wcv->procs++;
    }
  if (wcv->channels0 == 2 && wcv->channels1 == 1)
    {
      wcv->proc = g_realloc (wcv->proc,
                                (wcv->procs + 1) * sizeof (WaveConvProccess));
      wcv->proc[wcv->procs].func_begin
                            = (WaveConvBeginFunc)wave_convert_channels_begin;
      wcv->proc[wcv->procs].func = (WaveConvFunc)wave_convert_channels;
      wcv->proc[wcv->procs].func_end
                                = (WaveConvEndFunc)wave_convert_channels_end;
      wcv->procs++;
    }
  for (i = 0; i < wcv->procs; i++)
    wcv->proc[i].handle = wcv->proc[i].func_begin (wfx0, wfx1);
  return wcv;
}


/*  ja:PCMの変換をする
            wcv,変換ハンドル
         in_buf,入力データ
     in_samples,入力データのサンプル数
        out_buf,出力データ
    out_samples,出力データのサンプル数
            RET,TRUE:正常終了,FALSE:エラー                                  */
gboolean
wave_convert (WaveConv      *wcv,
              gpointer       in_buf,
              const gsize    in_samples,
              gpointer      *out_buf,
              gsize         *out_samples)
{
  gpointer buf;
  gsize samples;
  gint i;

  buf = in_buf;
  samples = in_samples;
  for (i = 0; i < wcv->procs; i++)
    {
      gpointer x_buf;

      wcv->proc[i].func (wcv->proc[i].handle, buf, samples, &x_buf, &samples);
      if (buf != in_buf)
        g_free (buf);
      buf = x_buf;
    }
  if (out_buf)
    *out_buf = buf;
  if (out_samples)
    *out_samples = samples;
  return TRUE;
}


/*  ja:PCMの変換を終了する
      wave_conv,変換ハンドル
        out_buf,出力データ
    out_samples,出力データのサンプル数
            RET,TRUE:正常終了,FALSE:エラー                                  */
gboolean
wave_convert_end (WaveConv *wcv,
                  gpointer *out_buf,
                  gsize    *out_samples)
{
  gpointer buf = NULL;
  gsize samples = 0, block_align;
  gint i;

  block_align = wcv->channels1 * wcv->bits_per_sample1 / 8;
  for (i = 0; i < wcv->procs; i++)
    {
      gpointer e_buf;
      gsize e_samples;
      gint j;

      wcv->proc[i].func_end (wcv->proc[i].handle, &e_buf, &e_samples);
      for (j = i + 1; j < wcv->procs; j++)
        {
          gpointer x_buf;

          wcv->proc[j].func (wcv->proc[j].handle,
                                        e_buf, e_samples, &x_buf, &e_samples);
          g_free (e_buf);
          e_buf = x_buf;
        }
      buf = g_realloc (buf, (samples + e_samples) * block_align);
      g_memmove ((guint8 *)buf + samples * block_align, e_buf,
                                                    e_samples * block_align);
      g_free (e_buf);
      samples += e_samples;
    }
  if (out_buf)
    *out_buf = buf;
  if (out_samples)
    *out_samples = samples;
  g_free (wcv->proc);
  g_free (wcv);
  return TRUE;
}


/*  ja:PCMの変換をする
           wfx0,元のWaveFormatEx構造体へのポインタ
           wfx1,新しいWaveFormatEx構造体へのポインタ
         in_buf,入力データ
     in_samples,入力データのサンプル数
        out_buf,出力データ
    out_samples,出力データのサンプル数
            RET,TRUE:正常終了,FALSE:エラー                                  */
gboolean
wave_convert_direct (WaveFormatEx  *wfx0,
                     WaveFormatEx  *wfx1,
                     gpointer       in_buf,
                     const gsize    in_samples,
                     gpointer      *out_buf,
                     gsize         *out_samples)
{
  gpointer buf, e_buf;
  gsize samples, e_samples, block_align;
  WaveConv *wcv;

  block_align = wfx_get_channels (wfx1) * wfx_get_bits_per_sample (wfx1) / 8;
  wcv = wave_convert_begin (wfx0, wfx1);
  wave_convert (wcv, in_buf, in_samples, &buf, &samples);
  wave_convert_end (wcv, &e_buf, &e_samples);
  buf = g_realloc (buf, (samples + e_samples) * block_align);
  g_memmove ((guint8 *)buf + samples * block_align, e_buf,
                                                    e_samples * block_align);
  g_free (e_buf);
  if (out_buf)
    *out_buf = buf;
  if (out_samples)
    *out_samples = samples;
  return TRUE;
}

==> vmaid-1.9.2.tar.gz: vmaid-1.9.2/avicore/wave.c <==
/*
    avicore
    copyright (c) 1998-2005 Kazuki IWAMOTO http://www.maid.org/ iwm@maid.org

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "wave.h"


/******************************************************************************
*                                                                             *
* ja:PCM加工関数                                                              *
*                                                                             *
******************************************************************************/
typedef struct _WaveConvChannels
{
  guint16 channels;
  guint16 bits_per_sample;
} WaveConvChannels;


/*  ja:PCMのチャンネル変換を開始する
    wfx0,元のWaveFormatEx構造体へのポインタ
    wfx1,新しいWaveFormatEx構造体へのポインタ
     RET,変換ハンドル,NULL:変換不可能                                       */
WaveConvChannels *
wave_convert_channels_begin (WaveFormatEx *wfx0,
                             WaveFormatEx *wfx1)
{
  WaveConvChannels *wcc;

  wcc = g_malloc (sizeof (WaveConvChannels));
  wcc->channels = wfx_get_channels (wfx0);
  wcc->bits_per_sample = wfx_get_bits_per_sample (wfx0);
  return wcc;
}


/*  ja:PCMのチャンネル変換をする
            wcc,変換ハンドル
         in_buf,入力データ
     in_samples,入力データのサンプル数
        out_buf,出力データ
    out_samples,出力データのサンプル数
            RET,TRUE:正常終了,FALSE:エラー                                  */
gboolean
wave_convert_channels (WaveConvChannels *wcc,
                       gpointer          in_buf,
                       const gsize       in_samples,
                       gpointer         *out_buf,
                       gsize            *out_samples)
{
  gpointer result;
  gsize i;

  if (wcc->channels == 1)
    {
      if (wcc->bits_per_sample == 8)
        {
          result = g_malloc (in_samples * 2 * sizeof (guint8));
          for (i = 0; i < in_samples; i++)
            ((guint8 *)result)[i * 2] =
            ((guint8 *)result)[i * 2 + 1] = ((guint8 *)in_buf)[i];
        }
      else
        {
          result = g_malloc (in_samples * 2 * sizeof (gint16));
          for (i = 0; i < in_samples; i++)
            ((gint16 *)result)[i * 2] =
            ((gint16 *)result)[i * 2 + 1] = ((gint16 *)in_buf)[i];
        }
    }
  else
    {
      if (wcc->bits_per_sample == 8)
        {
          result = g_malloc (in_samples * sizeof (guint8));
          for (i = 0; i < in_samples; i++)
            ((guint8 *)result)[i] = (((guint8 *)in_buf)[i * 2]
                                        + ((guint8 *)in_buf)[i * 2 + 1]) / 2;
        }
      else
        {
          result = g_malloc (in_samples * sizeof (gint16));
          for (i = 0; i < in_samples; i++)
            ((gint16 *)result)[i] =
                    GINT16_TO_LE ((GINT16_FROM_LE (((gint16 *)in_buf)[i * 2])
                        + GINT16_FROM_LE (((gint16 *)in_buf)[i * 2 + 1])) / 2);
        }
    }
  if (out_buf)
    *out_buf = result;
  else
    g_free (result);
  if (out_samples)
    *out_samples = in_samples;
  return TRUE;
}


/*  ja:PCMのチャンネル変換を終了する
      wave_conv,変換ハンドル
        out_buf,出力データ
    out_samples,出力データのサンプル数
            RET,TRUE:正常終了,FALSE:エラー                                  */
gboolean
wave_convert_channels_end (WaveConvChannels *wcc,
                           gpointer         *out_buf,
                           gsize            *out_samples)
{
  if (out_buf)
    *out_buf = NULL;
  if (out_samples)
    *out_samples = 0;
  g_free (wcc);
  return TRUE;
}


typedef struct _WaveConvBitPerSample
{
  guint16 channels;
  guint16 bits_per_sample;
} WaveConvBitPerSample;


/*  ja:PCMのビット数変換を開始する
    wfx0,元のWaveFormatEx構造体へのポインタ
    wfx1,新しいWaveFormatEx構造体へのポインタ
     RET,変換ハンドル,NULL:変換不可能                                       */
WaveConvBitPerSample *
wave_convert_bits_per_sample_begin (WaveFormatEx *wfx0,
                                    WaveFormatEx *wfx1)
{
  WaveConvBitPerSample *wcb;

  wcb = g_malloc (sizeof (WaveConvBitPerSample));
  wcb->channels = wfx_get_channels (wfx0);
  wcb->bits_per_sample = wfx_get_bits_per_sample (wfx0);
  return wcb;
}


/*  ja:PCMのビット数変換をする
            wcb,変換ハンドル
         in_buf,入力データ
     in_samples,入力データのサンプル数
        out_buf,出力データ
    out_samples,出力データのサンプル数
            RET,TRUE:正常終了,FALSE:エラー                                  */
gboolean
wave_convert_bits_per_sample (WaveConvBitPerSample *wcb,
                              gpointer              in_buf,
                              const gsize           in_samples,
                              gpointer             *out_buf,
                              gsize                *out_samples)
{
  gpointer result;
  gint i;

  if (wcb->bits_per_sample == 8)
    {
      result = g_malloc (in_samples * wcb->channels * sizeof (gint16));
      for (i = in_samples * wcb->channels - 1; i >= 0; i--)
        {
          gint c;

          c = ((guint8 *)in_buf)[i] - 128;
          c = c * 256 + (c % 2 == 0 ? 0 : 255);
          ((gint16 *)result)[i] = GINT16_TO_LE (c);
        }
    }
  else
    {
      result = g_malloc (in_samples * wcb->channels * sizeof (guint8));
      for (i = in_samples * wcb->channels - 1; i >= 0; i--)
        ((guint8 *)result)[i] = (GINT16_FROM_LE (((gint16 *)in_buf)[i])
                                                                + 32768) / 256;
    }
  if (out_buf)
    *out_buf = result;
  else
    g_free (result);
  if (out_samples)
    *out_samples = in_samples;
  return TRUE;
}


/*  ja:PCMのビット数変換を終了する
      wave_conv,変換ハンドル
        out_buf,出力データ
    out_samples,出力データのサンプル数
            RET,TRUE:正常終了,FALSE:エラー                                  */
gboolean
wave_convert_bits_per_sample_end (WaveConvBitPerSample *wcb,
                                  gpointer             *out_buf,
                                  gsize                *out_samples)
{
  if (out_buf)
    *out_buf = NULL;
  if (out_samples)
    *out_samples = 0;
  g_free (wcb);
  return TRUE;
}


typedef struct _WaveConvGeneric *WaveConvGeneric;
typedef WaveConvGeneric (*WaveConvBeginFunc)(WaveFormatEx *wfx0,
                                             WaveFormatEx *wfx1);
typedef gboolean (*WaveConvFunc)(WaveConvGeneric  wch,
                                 gpointer         in_buf,
                                 gsize            in_samples,
                                 gpointer        *out_buf,
                                 gsize           *out_samples);
typedef gboolean (*WaveConvEndFunc)(WaveConvGeneric  wch,
                                    gpointer        *out_buf,
                                    gsize           *out_samples);
typedef struct _WaveConvProccess
{
  WaveConvBeginFunc func_begin;
  WaveConvFunc func;
  WaveConvEndFunc func_end;
  WaveConvGeneric handle;
} WaveConvProccess;
struct _WaveConv
{
  guint16 channels0;
  guint16 channels1;
  guint16 bits_per_sample0;
  guint16 bits_per_sample1;
  gint procs;
  WaveConvProccess *proc;
};


/*  ja:PCMの変換を開始する
    wfx0,元のWaveFormatEx構造体へのポインタ
    wfx1,新しいWaveFormatEx構造体へのポインタ
     RET,変換ハンドル,NULL:変換不可能                                       */
WaveConv *
wave_convert_begin (WaveFormatEx *wfx0,
                    WaveFormatEx *wfx1)
{
  gint i;
  WaveConv *wcv;
  WaveFormatEx *wfx;

  wcv = g_malloc0 (sizeof (WaveConv));
  wcv->channels0 = wfx_get_channels (wfx0);
  wcv->channels1 = wfx_get_channels (wfx1);
  wcv->bits_per_sample0 = wfx_get_bits_per_sample (wfx0);
  wcv->bits_per_sample1 = wfx_get_bits_per_sample (wfx1);
  wfx = g_memdup (wfx0, wf_header_bytes (wfx0));
  if (wcv->bits_per_sample0 == 8 && wcv->bits_per_sample1 == 16)
    {
      wcv->proc = g_realloc (wcv->proc,
                                (wcv->procs + 1) * sizeof (WaveConvProccess));
      wcv->proc[wcv->procs].func_begin
                    = (WaveConvBeginFunc)wave_convert_bits_per_sample_begin;
      wcv->proc[wcv->procs].func = (WaveConvFunc)wave_convert_bits_per_sample;
      wcv->proc[wcv->procs].func_end
                        = (WaveConvEndFunc)wave_convert_bits_per_sample_end;
      wcv->proc[wcv->procs].handle
                                = wcv->proc[wcv->procs].func_begin (wfx, wfx1);
      if (!wcv->proc[wcv->procs].handle)
        {
          for (i = wcv->procs - 1; i >= 0; i--)
            wcv->proc[i].func_end (wcv->proc[i].handle, NULL, NULL);
          g_free (wcv);
          g_free (wfx);
          return NULL;
        }
      wfx_set_bits_per_sample (wfx, 16);
      wcv->procs++;
    }
  if (wcv->channels0 != wcv->channels1)
    {
      wcv->proc = g_realloc (wcv->proc,
                                (wcv->procs + 1) * sizeof (WaveConvProccess));
      wcv->proc[wcv->procs].func_begin
                            = (WaveConvBeginFunc)wave_convert_channels_begin;
      wcv->proc[wcv->procs].func = (WaveConvFunc)wave_convert_channels;
      wcv->proc[wcv->procs].func_end
                                = (WaveConvEndFunc)wave_convert_channels_end;
      wcv->proc[wcv->procs].handle
                                = wcv->proc[wcv->procs].func_begin (wfx, wfx1);
      if (!wcv->proc[wcv->procs].handle)
        {
          for (i = wcv->procs - 1; i >= 0; i--)
            wcv->proc[i].func_end (wcv->proc[i].handle, NULL, NULL);
          g_free (wcv);
          g_free (wfx);
          return NULL;
        }
      wfx_set_channels (wfx, wcv->channels1);
      wcv->procs++;
    }
  if (wcv->bits_per_sample0 == 16 && wcv->bits_per_sample1 == 8)
    {
      wcv->proc = g_realloc (wcv->proc,
                                (wcv->procs + 1) * sizeof (WaveConvProccess));
      wcv->proc[wcv->procs].func_begin
                    = (WaveConvBeginFunc)wave_convert_bits_per_sample_begin;
      wcv->proc[wcv->procs].func = (WaveConvFunc)wave_convert_bits_per_sample;
      wcv->proc[wcv->procs].func_end
                        = (WaveConvEndFunc)wave_convert_bits_per_sample_end;
      wcv->proc[wcv->procs].handle
                                = wcv->proc[wcv->procs].func_begin (wfx, wfx1);
      if (!wcv->proc[wcv->procs].handle)
        {
          for (i = wcv->procs - 1; i >= 0; i--)
            wcv->proc[i].func_end (wcv->proc[i].handle, NULL, NULL);
          g_free (wcv);
          g_free (wfx);
          return NULL;
        }
      wfx_set_bits_per_sample (wfx, 8);
      wcv->procs++;
    }
  g_free (wfx);
  return wcv;
}


/*  ja:PCMの変換をする
            wcv,変換ハンドル
         in_buf,入力データ
     in_samples,入力データのサンプル数
        out_buf,出力データ
    out_samples,出力データのサンプル数
            RET,TRUE:正常終了,FALSE:エラー                                  */
gboolean
wave_convert (WaveConv      *wcv,
              gpointer       in_buf,
              const gsize    in_samples,
              gpointer      *out_buf,
              gsize         *out_samples)
{
  gpointer buf;
  gsize samples, block_align;
  gint i;

  block_align = wcv->channels0 * wcv->bits_per_sample0 / 8;
  buf = g_memdup (in_buf, in_samples * block_align);
  samples = in_samples;
  for (i = 0; i < wcv->procs; i++)
    {
      gpointer x_buf;

      if (!wcv->proc[i].func (wcv->proc[i].handle,
                                            buf, samples, &x_buf, &samples))
        {
          g_free (buf);
          return FALSE;
        }
      g_free (buf);
      buf = x_buf;
    }
  if (out_buf)
    *out_buf = buf;
  else
    g_free (buf);
  if (out_samples)
    *out_samples = samples;
  return TRUE;
}


/*  ja:PCMの変換を終了する
      wave_conv,変換ハンドル
        out_buf,出力データ
    out_samples,出力データのサンプル数
            RET,TRUE:正常終了,FALSE:エラー                                  */
gboolean
wave_convert_end (WaveConv *wcv,
                  gpointer *out_buf,
                  gsize    *out_samples)
{
  gboolean result = TRUE;
  gpointer buf = NULL;
  gsize samples = 0, block_align;
  gint i;

  block_align = wcv->channels1 * wcv->bits_per_sample1 / 8;
  for (i = 0; i < wcv->procs; i++)
    {
      gpointer e_buf;
      gsize e_samples;
      gint j;

      if (!wcv->proc[i].func_end (wcv->proc[i].handle, &e_buf, &e_samples))
        result = FALSE;
      if (!result)
        continue;
      for (j = i + 1; j < wcv->procs; j++)
        {
          gpointer x_buf;

          wcv->proc[j].func (wcv->proc[j].handle,
                                        e_buf, e_samples, &x_buf, &e_samples);
          g_free (e_buf);
          e_buf = x_buf;
        }
      buf = g_realloc (buf, (samples + e_samples) * block_align);
      g_memmove ((guint8 *)buf + samples * block_align, e_buf,
                                                    e_samples * block_align);
      g_free (e_buf);
      samples += e_samples;
    }
  if (result && out_buf)
    *out_buf = buf;
  else
    g_free (buf);
  if (result && out_samples)
    *out_samples = samples;
  g_free (wcv->proc);
  g_free (wcv);
  return result;
}


/*  ja:PCMの変換をする
           wfx0,元のWaveFormatEx構造体へのポインタ
           wfx1,新しいWaveFormatEx構造体へのポインタ
         in_buf,入力データ
     in_samples,入力データのサンプル数
        out_buf,出力データ
    out_samples,出力データのサンプル数
            RET,TRUE:正常終了,FALSE:エラー                                  */
gboolean
wave_convert_direct (WaveFormatEx  *wfx0,
                     WaveFormatEx  *wfx1,
                     gpointer       in_buf,
                     const gsize    in_samples,
                     gpointer      *out_buf,
                     gsize         *out_samples)
{
  gpointer buf, e_buf;
  gsize samples, e_samples, block_align;
  WaveConv *wcv;

  block_align = wfx_get_channels (wfx1) * wfx_get_bits_per_sample (wfx1) / 8;
  wcv = wave_convert_begin (wfx0, wfx1);
  wave_convert (wcv, in_buf, in_samples, &buf, &samples);
  wave_convert_end (wcv, &e_buf, &e_samples);
  buf = g_realloc (buf, (samples + e_samples) * block_align);
  g_memmove ((guint8 *)buf + samples * block_align, e_buf,
                                                    e_samples * block_align);
  g_free (e_buf);
  if (out_buf)
    *out_buf = buf;
  if (out_samples)
    *out_samples = samples;
  return TRUE;
}
